/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef ECC_GPU_CUH_
#define ECC_GPU_CUH_

#include <string>

#include "utils.hpp"
#include "integer.cuh"
#include "gfp_gpu.cuh"

#ifdef USE_NAF
#include "wnaf.cuh"
#endif

// shared memory space calculator
template<uint N, uint NBVAR> 
struct ECCSharedMem {enum {Value=NB_LIMB*NBVAR};};


// memory plan for point allocation in local memory (may be shared)
struct PointLocalMemory{
	typedef LocalMemory AllocX;
	typedef LocalMemory AllocY;
	typedef LocalMemory AllocZ;
};

// memory plan for point allocation in texture memory at position k
template<uint k>
struct PointTextureMemory {
	typedef TextureMemory<k>   AllocX;
	typedef TextureMemory<k+1> AllocY;
	typedef TextureMemory<k+2> AllocZ;
};


//Class point<N> : Point = (_X,_Y,_Z)
template <uint N, typename PointAlloc=PointLocalMemory>
class Point{
public:
	Integer<N,typename PointAlloc::AllocX> _x;
	Integer<N,typename PointAlloc::AllocY> _y; 
	Integer<N,typename PointAlloc::AllocZ> _z;

	//default constructor
	_HD_ Point(); 

	//constructor of the point using 3 string representing the 3 coordinates
	_H_  Point(const std::string&,const std::string&,const std::string&);	

	//assignment operator (needed for scalarMul)
	template <typename AllocPoint2>
	_HD_ Point& operator=(const Point<N, AllocPoint2>&);
}; 

template <uint N, typename AllocGFP=LocalMemory, typename AllocA=LocalMemory>
class ECurve;


//template<uint N, typename AllocGFP=LocalMemory, typename AllocA=LocalMemory>
template<uint N, typename AllocGFP, typename AllocA>
class ECurve {
public:
	typedef AllocGFP MemoryGFP;
	typedef AllocA   MemoryA;

	//Parameter a of the curve
#ifdef __UseTex
	const Integer<N, AllocA>    _a;
#else
	Integer<N, AllocA> _a;
#endif
	
	//Primary field in which the curve is defined
	GFp<N, AllocGFP>     _Fp;	

	// pointer to shared memory for extra variables
	Integer<N, LocalMemory> _var1;
	Integer<N, LocalMemory> _var2;
	Integer<N, LocalMemory> _var3;

	// NAF 
#ifdef USE_NAF
	Integer<N, LocalMemory> _var4;
#ifdef NAF3
	Point<N>        _P3;
#endif
#ifdef OWNAF
	OWNaf<N,NAF_W> _naf;
#else
	Naf<N> _naf;
#endif
#endif

	// default constructor
	_HD_ ECurve();

	// constructor of the curve from string parameters a  and the prime (b is not needed)
	_H_  ECurve(string, string);

	// addition of 2 points on the curve
	template<typename AllocPoint1, typename AllocPoint2>
	_HD_ void add_in(Point<N,AllocPoint1>&, const Point<N,AllocPoint2>&);

	// addition of 2 points on the curve - version for non optimized square in Fp
	template<typename AllocPoint1, typename AllocPoint2>
	_HD_ void add_in2(Point<N,AllocPoint1>&, const Point<N,AllocPoint2>&);
	
	// doubling of a point on the curve
	template <typename AllocPoint1>
	_HD_ void dbl_in(Point<N, AllocPoint1>&);

	// doubling of a point on the curve - version for non optimized square in Fp
	template <typename AllocPoint1>
	_HD_ void dbl_in2(Point<N, AllocPoint1>&);

	//compute Q = 2Q+P
	template<typename AllocPoint1, typename AllocPoint2>
	_HD_ void dbl_add_in(Point<N,AllocPoint1>&, const Point<N,AllocPoint2>&);
	
	// scalar multiplication on the curve
	template<typename Alloc, typename AllocPoint1, typename AllocPoint2>
	_D_ void scalarMulLeftToRight(Point<N,AllocPoint1>&, const Integer<N,Alloc>&, const Point<N, AllocPoint2>&); 

	// scalar multiplication on the curve
	template<typename Alloc, typename AllocPoint1, typename AllocPoint2>
	_D_ void scalarMulRightToLeft(Point<N,AllocPoint1>&, const Integer<N,Alloc>&, Point<N, AllocPoint2>&); 

#ifdef USE_NAF
	template<typename AllocPoint1, typename AllocPoint2>
	_HD_ void sub_in2(Point<N,AllocPoint1>&, const Point<N,AllocPoint2>&);

	template<typename AllocPoint1>
	_HD_ void smul_in(Point<N, AllocPoint1>&, int);

	// scalar multiplication on the curve with NAF
	template<typename Alloc, typename AllocPoint1, typename AllocPoint2>
	_D_ void scalarMulLeftToRightNaf(Point<N,AllocPoint1>&, const Integer<N,Alloc>&, const Point<N, AllocPoint2>&); 
#endif

};

#include "ecc_gpu.cu"

#endif //ECC_GPU_CUH_
