/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */


/************************************************************************************************/
//GFp FUNCTIONS

//default constructor
template<uint N, typename Alloc>
_HD_ GFp<N,Alloc>::GFp(): _prime() {}

//construct GFP from a string (the prime) and compute -p^-1 % 2^16 (needed for montgomery multiplication)
template<uint N, typename Alloc>
_H_ GFp<N,Alloc>::GFp(string s) : _prime(s) {
	_invp  = Utils::modulo_inverse((1 << 16), (-_prime[0]& 0xFFFF));
}

/************************************************************************************************/

//FIXME - in this addition, the size of the result can be greather than its real size

//Modular addition : res = (res+op1) % prime

template <uint N, typename Alloc>
template<typename Alloc1, typename Alloc2>
_D_ void GFp<N,Alloc>::add_in(Integer<N,Alloc1>& res, const Integer<N,Alloc2>& op1) {
	uint carry = 0;
	//Add the two number with carry propagation
#pragma unroll
	for(int i = 0; i < NB_LIMB; i++) {
		uint add = op1[i]+ res[i]+carry;
		res[i] = add & 0xFFFF;
		carry = add >> 16;
	}
	//if the result is superior to the prime
	if(res >= _prime || carry) {
		carry = 0;
#pragma unroll
		for(int i = 0; i < NB_LIMB; i++) {
			int sub = res[i]-_prime.get(i)-carry;//_prime[i]-carry;
			carry = (sub < 0 ? 1 : 0);
			res[i] = sub & 0xFFFF;
		}
	}
}


/************************************************************************************************/

//FIXME - in this addition, the size of the result can be greather than its real size

//Modular addition
template<uint N, typename Alloc>
template<typename Alloc1, typename Alloc2, typename Alloc3>
_D_ void GFp<N,Alloc>::add(Integer<N,Alloc1>& res, const Integer<N,Alloc2>& op1, const Integer<N,Alloc3>& op2) const{
	uint carry = 0;
	//Add the two number with carry propagation
#pragma unroll
	for(int i = 0; i < NB_LIMB; i++) {
		uint add = op1[i]+ op2[i]+carry;
		res[i] = add & 0xFFFF;
		carry = add >> 16;
	}
	if(res >= _prime || carry) {
		carry = 0;
#pragma unroll
		for(int i = 0; i < NB_LIMB; i++) {
			int sub = res[i]-_prime[i]-carry;
			carry = (sub < 0 ? 1 : 0);
			res[i] = sub & 0xFFFF;
		}
	}
}

/************************************************************************************************/

//FIXME - in the subtraction, the size of the result can be greather than its real size

//Modular subtraction : res = (res-op1) % prime
template <uint N, typename Alloc>
template<typename Alloc1, typename Alloc2>
_D_ void GFp<N,Alloc>::sub_in(Integer<N,Alloc1>& res, const Integer<N,Alloc2>& op1) {
	if(res >= op1) {
		uint carry = 0;
#pragma unroll
		for(int i = 0; i < NB_LIMB; i++) {
			int sub = res[i]-op1[i]-carry;
			carry = (sub < 0 ? 1 : 0);
			res[i] = sub & (0xFFFF);
		}
	}
	else {
		uint carry = 0;
#pragma unroll
		for(int i = 0; i < NB_LIMB; i++) {
			int sub = res[i]-op1[i]+carry+_prime.get(i);
			carry = (sub < 0 ? -1 : (sub >= (1 << 16) ? 1 : 0));
			res[i] = sub & 0xFFFF;
		}
	}
}

/************************************************************************************************/

//FIXME - in the subtraction, the size of the result can be greather than its real size

//Modular subtraction
template<uint N, typename Alloc>
template<typename Alloc1, typename Alloc2, typename Alloc3>
_D_ void GFp<N,Alloc>::sub(Integer<N,Alloc1>& res, const Integer<N,Alloc2>& op1, const Integer<N,Alloc3>& op2) const{
	if(op1 >= op2) {
		uint carry = 0;
#pragma unroll
		for(int i = 0; i < NB_LIMB; i++) {
			int sub = op1[i]-op2[i]-carry;
			carry = (sub < 0 ? 1 : 0);
			res[i] = sub & (0xFFFF);
		}
	}
	else {
		uint carry = 0;
#pragma unroll
		for(int i = 0; i < NB_LIMB; i++) {
			int sub = op1[i]-op2[i]+carry+_prime[i];
			carry = (sub < 0 ? -1 : (sub >= (1 << 16) ? 1 : 0));
			res[i] = sub & 0xFFFF;
		}
	}
}

/************************************************************************************************/


//Montgomery Multiplication between to Integers : compute op1*op2*R % prime where R = 2^N
template<uint N, typename Alloc>
template<typename Alloc1, typename Alloc2, typename Alloc3>
_D_ void GFp<N,Alloc>::mul(Integer<N,Alloc1>& resp, const Integer<N,Alloc2>& op1, const Integer<N,Alloc3>& op2) {

	_var1.init(0);
	//Store the carry of the addition (Res + ui * op2 + x[i] * _prime)/b
	uint carryA = 0;
	uint y0 = op2[0];
#pragma unroll
	for(int i = 0; i < NB_LIMB; i++) {

		//xi 
		uint xi = op1[i];

		// xiy =  xi * y0
		uint xiy = __umul24(xi,y0);
		
		//ui = res[0] + xi*y0 mod 2^16 
		uint ui = _var1[0]+xiy;

		//ui = (ui*invp) mod 2^16
		ui = __umul24(ui,_invp);
		ui = ui & 0xFFFF;

		//c = ui * prime[0]
		uint c = __umul24(ui,_prime[0]);
		

		//carry = (xi*y0 + ui*prime[0]) / 2^16
		uint carry = xiy >> 16;  ;
		carry += (c >> 16);

		//c = res[0] + xi*y0 + ui*prime[0]   mod 2^16
		c = c & 0xFFFF;
		c += (xiy & 0xFFFF) + _var1[0];

		// carry += c / 2^16
		carry += (c >> 16);

       		// compute (res+ui * prime + x[i] * y)/b.  The carry is stored in carryA
#pragma unroll
		for(int j = 1; j < NB_LIMB; j++) {
						
			uint r = _var1[j] + carry;
			carry = (r >> 16);
			
			r = r & 0xFFFF;
			//c = op2 * x[i]
			c =  __umul24(xi, op2[j]);
			carry += (c >> 16);
			r += (c & 0xFFFF);
			//c = ui * prime[j]
			c = __umul24(ui, _prime.get(j));
			carry += (c >> 16);
			r += (c & 0xFFFF);
			carry += (r >> 16);
			_var1[j-1] = (r & 0xFFFF);
		}
		_var1[NB_LIMB-1] = ((carry) + carryA) & 0xFFFF;
		carryA = (carry >> 16);
	}
	//FIXME - Size can be greather than the real size of res
	//If the last addition has a carry or the result is greather than the prime, subtract the prime
	if(carryA || _var1 >= _prime) {
		uint carry = 0;
#pragma unroll
		for(int i = 0; i < NB_LIMB;i++) {
			int sub = _var1[i]-_prime.get(i)-carry;
			carry = (sub < 0 ? 1 : 0);
			_var1[i] = sub & 0xFFFF;
		}
	}
	resp = _var1;
}

/************************************************************************************************/


//multiply one integer op2 by one int op1 : op1*op2 % prime
template<uint N, typename Alloc>
template<typename Alloc1, typename Alloc2>
_D_ void GFp<N,Alloc>::mul(Integer<N,Alloc1>& res, int op1, const Integer<N,Alloc2>& op2) {

	//Multiply each mp_d by the int
	uint carry = 0;
#pragma unroll
	for(int i = 0; i < NB_LIMB; i++) {
		int mul = __umul24(abs(op1), op2[i])+carry;
		res[i] = mul & 0xFFFF;
		carry = mul >> 16;
	}
	uint m_t = _prime.get(NB_LIMB-1);
	uint m_t1 = _prime.get(NB_LIMB-2);
	//compute the partial quotient q= op1*op2 / m
	int q = ((carry << 16)+res[NB_LIMB-1])/m_t;

	//decrease the quotient while q*m > op1*op2
	ulong mul = __umul24(q,m_t1);
	uint h_p = q*(m_t)+(mul >> 16);
	uint b_p = mul & 0xFFFF;
	while(h_p > ((carry << 16) + res[NB_LIMB-1]) || ((h_p == ((carry << 16) + res[NB_LIMB-1]) && b_p > res[NB_LIMB-2]) )) {
		q--;
		mul = __umul24(q,m_t1);
		h_p = __umul24(q,m_t)+(mul >> 16);
		b_p = mul & 0xFFFF;
	}
	
	uint car = 0;
	uint c = 0;
	//subtraction op1*op2 - q*m
#pragma unroll
	for(int i = 0; i < NB_LIMB; i++) {
		mul = __umul24(q,_prime.get(i))+car;
	        b_p = mul & 0xFFFF;
		car = mul >> 16;
		
		int sub = res[i] - b_p -c;
		if(sub >= 0) {
			res[i] = sub;
			c = 0;
		}
		else {
			res[i] = sub & 0xFFFF;
			c = 1;
		}
	}
	if(op1 < 0) {
		sub(res, _prime, res);
	}
}

/************************************************************************************************/

//Compute op*R^-1 mod m
template<uint N, typename Alloc>
template<typename Alloc1, typename Alloc2>
_D_ void GFp<N,Alloc>::convertFromMontg(Integer<N,Alloc1>& res, const Integer<N,Alloc2>& op) {
	_var1 = op;
	//at each step, _var1 = 0 so we can divide by b
#pragma unroll
	for(int i = 0; i < NB_LIMB; i++) {
		//compute (_var1[0] * _invp) % b
		uint ui = __umul24(_var1[0],_invp) & 0xFFFF;
		//_var1[0] + _prime[0]
		uint add = _var1[0] +__umul24(ui,_prime[0]);
		//add % b = 0
		uint carry = add >> 16;
		for(int j = 1; j < NB_LIMB+1; j++) {
			//compute _var1[i] = _var1[i+1] +ui*m[i];
			if(j != NB_LIMB) add  = _var1[j]+__umul24(ui, _prime[j])+carry;
			else add = carry;
			_var1[j-1] = add & 0xFFFF;
			carry = add >> 16;
		}
	}
	if(_var1 >= _prime) {
		sub(_var1,_var1,_prime);
	}
	res = _var1;
}

/************************************************************************************************/

//Compute op*R mod _prime with R = 2^(16*(N/4))
template<uint N, typename Alloc>
template<typename Alloc1, typename Alloc2>
_D_ void GFp<N,Alloc>::convertToMontg(Integer<N,Alloc1>& res, const Integer<N,Alloc2>& op) {
	_var1 = op;
	uint m_s = NB_LIMB-1;
	uint m_t  = _prime[m_s];
	uint m_t1 = _prime[m_s-1];
	uint q;

	//at each step, the partial result  x is shift in size of the base, and we compute x = x*2^b - q*m
#pragma unroll
	for(int i = NB_LIMB-1; i>= 0; i--) {
		//Compute q = x*2^b / m 
		uint xi  = _var1[m_s]; 
		uint xi1 = _var1[m_s-1];
		uint xi2 = _var1[m_s-2];
		if(xi == m_t) q = 0xFFFF;
		else q = (uint)((xi << 16) + xi1)/m_t;

		uint mul = __umul24(q, m_t1);
		uint h_p = q*(m_t)+(mul >> 16);
		uint b_p = mul & 0xFFFF;
		while(h_p > ((xi << 16) + xi1) || ((h_p == ((xi << 16) + xi1) && b_p > xi2) )) {
			q--;
			mul = __umul24(q,m_t1);
			h_p = __umul24(q,m_t)+(mul >> 16);
			b_p = mul & 0xFFFF;
		}
	
		//compute m*q
		uint carry = 0;

		mul = __umul24(q,_prime[0])+carry;

		uint lsw = mul & 0xFFFF;
		carry = mul >> 16;

		//x*2^16 - m*q
		lsw = (1 << 16)-lsw;

		uint c = (op[i] == 0 ? 0 : 1);
		//x = x - q * prime
#pragma unroll
		for(int j = 1; j < NB_LIMB; j++) {
			mul = __umul24(q,_prime[j])+carry;
			carry = mul >> 16;

			int sub = _var1[j-1] - (mul & 0xFFFF) - c;
			if(sub >= 0) {
				_var1[j-1] = sub;
				c = 0;
			}
			else {
				_var1[j-1] = sub & 0xFFFF;
				c = 1;
			}
		}
		#pragma unroll
		for(uint j = NB_LIMB-1; j >= 1; j--) _var1[j] = _var1[j-1];
		_var1[0] = lsw;
	}
	res = _var1;



	/*	_var1 = op;
	uint m_s = NB_LIMB-1;
	uint m_t  = _prime[m_s];
	uint m_t1 = _prime[m_s-1];
	uint q;

	//at each step, the partial result  x is shift in size of the base, and we compute x = x*2^b - q*m
#pragma unroll
	for(int i = NB_LIMB-1; i>= 0; i--) {
		//Compute q = x*2^b / m 
		uint xi  = _var1[m_s]; 
		uint xi1 = _var1[m_s-1];
		uint xi2 = _var1[m_s-2];
		if(xi == m_t) q = 0xFFFF;
		else q = (uint)((xi << 16) + xi1)/m_t;

		uint mul = __umul24(q, m_t1);
		uint h_p = q*(m_t)+(mul >> 16);
		uint b_p = mul & 0xFFFF;
		while(h_p > ((xi << 16) + xi1) || ((h_p == ((xi << 16) + xi1) && b_p > xi2) )) {
			q--;
			mul = __umul24(q,m_t1);
			h_p = __umul24(q,m_t)+(mul >> 16);
			b_p = mul & 0xFFFF;
		}
	
		//compute m*q
		uint carry = 0;
		//		Integer<N> xt;
		_var2.init(0);
		mul = __umul24(q,_prime[0])+carry;
		_var2[0] = mul & 0xFFFF;
		carry = mul >> 16;

		//x*2^16 - m*q
		_var2[0] = (1 << 16)-_var2[0];

		uint c = (op[i] == 0 ? 0 : 1);
		//x = x - q * prime
#pragma unroll
		for(int j = 1; j < NB_LIMB; j++) {
			mul = __umul24(q,_prime[j])+carry;
			_var2[j] = mul & 0xFFFF;
			carry = mul >> 16;

			int sub = _var1[j-1] - _var2[j]-c;
			if(sub >= 0) {
				_var2[j] = sub;
				c = 0;
			}
			else {
				_var2[j] = sub & 0xFFFF;
				c = 1;
			}
		}
		_var1 = _var2;
	}
	res = _var1;*/
}

/************************************************************************************************/
