/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef KERNEL_EC_ARITHMETIC_CUH_
#define KERNEL_EC_ARITHMETIC_CUH_

#include "gpu_memory.cuh"
#include "integer.cuh"
#include "integer-reg.cuh"
#include "ecc_gpu.cuh"


/*************************
 * SHARED MEMORY VERSION *
 *************************/

template<uint N>
struct SharedData4 {
	ECurve<N> _ell;
	Point<N>  _P;
	Point<N>  _Q;
};

template<uint N>
struct SharedData5 {
	ECurve<N> _ell;
	Point<N>  _P;
};

template <uint N>
struct SharedData {
	ECurve<N>     _ell;
	Integer<N>      _k;
	Point<N>        _P;
	Point<N>        _Q;
};

/***** ADDITION OF POINTS **************************************************************************************/

template <uint N, uint nb_op>
__global__ void ec_addition_shared(GPU_vector_data<N> data) {

	__shared__ uint shared_integer[BLOCK_SIZE*(sizeof(SharedData4<N>) >> 2)];

	SharedData4<N>* ptr = (SharedData4<N>*) &shared_integer[Utils::tid_block()*(sizeof(SharedData4<N>) >> 2)];
	
	// load the points P and Q from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._x, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._y, data.__v_Py, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._z, data.__v_Pz, Utils::tid());

	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_Q._x, data.__v_Qx, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_Q._y, data.__v_Qy, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_Q._z, data.__v_Qz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_ell._Fp._prime, data.__v_prime, Utils::tid());
	ptr->_ell._Fp._invp= data.__v_invp[Utils::tid()];
	
	//Point coordinate in Montgomery form (x*R mod prime ...)
	ptr->_ell._Fp.convertToMontg(ptr->_P._x, ptr->_P._x);
	ptr->_ell._Fp.convertToMontg(ptr->_P._y, ptr->_P._y);
	ptr->_ell._Fp.convertToMontg(ptr->_P._z, ptr->_P._z);
	ptr->_ell._Fp.convertToMontg(ptr->_Q._x, ptr->_Q._x);
	ptr->_ell._Fp.convertToMontg(ptr->_Q._y, ptr->_Q._y);
	ptr->_ell._Fp.convertToMontg(ptr->_Q._z, ptr->_Q._z);

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		ptr->_ell.add_in2(ptr->_Q, ptr->_P);
	}

	ptr->_ell._Fp.convertFromMontg(ptr->_Q._x, ptr->_Q._x);
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._y, ptr->_Q._y);
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._z, ptr->_Q._z);

	// store the Point Q  in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, ptr->_Q._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Py, ptr->_Q._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Pz, ptr->_Q._z, Utils::tid());
}

/***** DOUBLING OF POINT ***********************************************************************************************/

template <uint N, uint nb_op>
__global__ void ec_doubling_shared(GPU_vector_data<N> data) {
	__shared__ uint shared_integer[BLOCK_SIZE*(sizeof(SharedData5<N>) >> 2)];
	SharedData5<N>* ptr = (SharedData5<N>*) &shared_integer[Utils::tid_block()*(sizeof(SharedData5<N>) >> 2)];
	// load the point P from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._x, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._y, data.__v_Py, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._z, data.__v_Pz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_ell._a,  data.__v_a, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_ell._Fp._prime, data.__v_prime, Utils::tid());
	ptr->_ell._Fp._invp= data.__v_invp[Utils::tid()];
	
	//Point coordinate in Montgomery form (x*R mod prime ...)
	ptr->_ell._Fp.convertToMontg(ptr->_P._x, ptr->_P._x);
	ptr->_ell._Fp.convertToMontg(ptr->_P._y, ptr->_P._y);
	ptr->_ell._Fp.convertToMontg(ptr->_P._z, ptr->_P._z);
	ptr->_ell._Fp.convertToMontg(ptr->_ell._a, ptr->_ell._a);

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		ptr->_ell.dbl_in2(ptr->_P);
	}

	ptr->_ell._Fp.convertFromMontg(ptr->_P._x, ptr->_P._x);
	ptr->_ell._Fp.convertFromMontg(ptr->_P._y, ptr->_P._y);
	ptr->_ell._Fp.convertFromMontg(ptr->_P._z, ptr->_P._z);

	// store the Point P in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, ptr->_P._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Py, ptr->_P._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Pz, ptr->_P._z, Utils::tid());
}


/***** SCALAR MULTIPLICATION OF POINT ***********************************************************************************************/

template <uint N>
__global__ void ec_multiplication_shared(GPU_vector_data<N> data) {

	__shared__ uint shared_integer[BLOCK_SIZE*(sizeof(SharedData<N>) >> 2)];
	SharedData<N>* ptr = (SharedData<N>*) &shared_integer[Utils::tid_block()*(sizeof(SharedData<N>) >> 2)];


	// load the point P from global memory 	
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._x, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._y, data.__v_Py, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._z, data.__v_Pz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_ell._a,  data.__v_a, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_ell._Fp._prime, data.__v_prime, Utils::tid());
	ptr->_ell._Fp._invp= data.__v_invp[Utils::tid()];
			
	// load the integer k from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_k, data.__v_k, Utils::tid());

	//Point coordinate in Montgomery form (x*R mod prime ...)
	ptr->_ell._Fp.convertToMontg(ptr->_P._x, ptr->_P._x);
	ptr->_ell._Fp.convertToMontg(ptr->_P._y, ptr->_P._y);
	ptr->_ell._Fp.convertToMontg(ptr->_P._z, ptr->_P._z);
	ptr->_ell._Fp.convertToMontg(ptr->_ell._a, ptr->_ell._a);
	
	//sclar multiplication
	ptr->_ell.scalarMulRightToLeft(ptr->_Q, ptr->_k, ptr->_P);

	//convert PMon = (xMon, yMon, zMon) in Fp form (x*R*R^-1 mod prime)
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._x, ptr->_Q._x);
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._y, ptr->_Q._y);
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._z, ptr->_Q._z);

	// store the Point Q  in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, ptr->_Q._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Py, ptr->_Q._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Pz, ptr->_Q._z, Utils::tid());
}

/************************
 * LOCAL MEMORY VERSION *
 ************************/


/***** ADDITION OF POINTS **************************************************************************************/

template <uint N, uint nb_op>
__global__ void ec_addition_local(GPU_vector_data<N> data) {

	Point<N> P,Q;
	ECurve<N> ell;
	

	
	// load the points P and Q from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(P._x, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(P._y, data.__v_Py, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(P._z, data.__v_Pz, Utils::tid());

	Memory<N,NUMBER_THREADS>::globalLoad(Q._x, data.__v_Qx, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(Q._y, data.__v_Qy, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(Q._z, data.__v_Qz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ell._Fp._prime, data.__v_prime, Utils::tid());
	ell._Fp._invp= data.__v_invp[Utils::tid()];
	
	//Point coordinate in Montgomery form (x*R mod prime ...)
	ell._Fp.convertToMontg(P._x, P._x);
	ell._Fp.convertToMontg(P._y, P._y);
	ell._Fp.convertToMontg(P._z, P._z);
	ell._Fp.convertToMontg(Q._x, Q._x);
	ell._Fp.convertToMontg(Q._y, Q._y);
	ell._Fp.convertToMontg(Q._z, Q._z);

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		ell.add_in2(Q, P);
	}

	ell._Fp.convertFromMontg(Q._x, Q._x);
	ell._Fp.convertFromMontg(Q._y, Q._y);
	ell._Fp.convertFromMontg(Q._z, Q._z);

	// store the Point Q  in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, Q._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Py, Q._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Pz, Q._z, Utils::tid());
}

/***** DOUBLING OF POINT ***********************************************************************************************/

template <uint N, uint nb_op>
__global__ void ec_doubling_local(GPU_vector_data<N> data) {

	Point<N> P;
	ECurve<N> ell;

	// load the point P from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(P._x, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(P._y, data.__v_Py, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(P._z, data.__v_Pz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ell._a,  data.__v_a, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ell._Fp._prime, data.__v_prime, Utils::tid());
	ell._Fp._invp= data.__v_invp[Utils::tid()];
	
	//Point coordinate in Montgomery form (x*R mod prime ...)
	ell._Fp.convertToMontg(P._x, P._x);
	ell._Fp.convertToMontg(P._y, P._y);
	ell._Fp.convertToMontg(P._z, P._z);
	ell._Fp.convertToMontg(ell._a, ell._a);

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		ell.dbl_in2(P);
	}

	ell._Fp.convertFromMontg(P._x, P._x);
	ell._Fp.convertFromMontg(P._y, P._y);
	ell._Fp.convertFromMontg(P._z, P._z);

	// store the Point P in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, P._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Py, P._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Pz, P._z, Utils::tid());
}

/***** SCALAR MULTIPLICATION OF POINT ***********************************************************************************************/

template <uint N>
__global__ void ec_multiplication_local(GPU_vector_data<N> data) {

	
	ECurve<N> ell;
	Point<N>  P,Q;
	Integer<N> k;

	// load the point P from global memory 	
	Memory<N,NUMBER_THREADS>::globalLoad(P._x, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(P._y, data.__v_Py, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(P._z, data.__v_Pz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ell._a,  data.__v_a, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ell._Fp._prime, data.__v_prime, Utils::tid());
	ell._Fp._invp= data.__v_invp[Utils::tid()];
			
	// load the integer k from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(k, data.__v_k, Utils::tid());

	//Point coordinate in Montgomery form (x*R mod prime ...)
	ell._Fp.convertToMontg(P._x, P._x);
	ell._Fp.convertToMontg(P._y, P._y);
	ell._Fp.convertToMontg(P._z, P._z);
	ell._Fp.convertToMontg(ell._a, ell._a);
	
	//sclar multiplication
	ell.scalarMulRightToLeft(Q, k, P);

	//convert PMon = (xMon, yMon, zMon) in Fp form (x*R*R^-1 mod prime)
	ell._Fp.convertFromMontg(Q._x, Q._x);
	ell._Fp.convertFromMontg(Q._y, Q._y);
	ell._Fp.convertFromMontg(Q._z, Q._z);

	// store the Point Q  in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, Q._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Py, Q._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Pz, Q._z, Utils::tid());
}

/***************************
 * REGISTER MEMORY VERSION *
 ***************************/


/***** ADDITION OF POINTS **************************************************************************************/

template <uint N, uint nb_op>
__global__ void ec_addition_register(GPU_vector_data<N> data) {

	const uint NWORDS = N/32+ (N%32?1:0);
	IntegerReg<NWORDS> prime, var1,var2,var3,var4;
	PointReg<NWORDS> P,Q;
	uint minvp;
	
	
	// load the points P and Q from global memory 
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._x, (uint*)data.__v_Px, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._y, (uint*)data.__v_Py, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._z, (uint*)data.__v_Pz, Utils::tid());

	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(Q._x, (uint*)data.__v_Qx, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(Q._y, (uint*)data.__v_Qy, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(Q._z, (uint*)data.__v_Qz, Utils::tid());
	
	// load  the prime of Fp from global memory 
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(prime, (uint*)data.__v_prime, Utils::tid());
	minvp= data.__v_invp[Utils::tid()];
	
	//Point coordinate in Montgomery form (x*R mod prime ...)
	//convertToMontg(P._x, P._x, prime, var1);
	//convertToMontg(P._y, P._y, prime, var1);
	//convertToMontg(P._z, P._z, prime, var1);
	//convertToMontg(Q._x, Q._x, prime, var1);
	//convertToMontg(Q._y, Q._y, prime, var1);
	//convertToMontg(Q._z, Q._z, prime, var1);

	for(uint i =0; i < nb_op; i++) {
		addin(Q, P, prime, minvp, var1, var2, var3, var4);
	}

	//var1.set_one();
	//convertFromMontg(Q._x, Q._x, prime, minvp, var1, var2);
	//convertFromMontg(Q._y, Q._y, prime, minvp, var1, var2);
	//convertFromMontg(Q._z, Q._z, prime, minvp, var1, var2);

	// store the Point Q  in global memory
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Px, Q._x, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Py, Q._y, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Pz, Q._z, Utils::tid());
}

/***** DOUBLING OF POINT ***********************************************************************************************/

template <uint N, uint nb_op>
__global__ void ec_doubling_register(GPU_vector_data<N> data) {

	const uint NWORDS = N/32+ (N%32?1:0);	
	IntegerReg<NWORDS> prime, a, var1,var2,var3;
	PointReg<NWORDS> P;
	uint minvp;
	
	
	// load the points P  from global memory 
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._x, (uint*)data.__v_Px, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._y, (uint*)data.__v_Py, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._z, (uint*)data.__v_Pz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(a, (uint*)data.__v_a, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(prime, (uint*)data.__v_prime, Utils::tid());
	minvp= data.__v_invp[Utils::tid()];
	
	//Point coordinate in Montgomery form (x*R mod prime ...)
	//convertToMontg(P._x, P._x, prime, var1);
	//convertToMontg(P._y, P._y, prime, var1);
	//convertToMontg(P._z, P._z, prime, var1);

	for(uint i =0; i < nb_op; i++) {
		dblin(P, a, prime, minvp, var1, var2, var3);
	}

	//var1.set_one();
	//convertFromMontg(P._x, P._x, prime, minvp, var1, var2);
	//convertFromMontg(P._y, P._y, prime, minvp, var1, var2);
	//convertFromMontg(P._z, P._z, prime, minvp, var1, var2);

	// store the Point Q  in global memory
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Px, P._x, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Py, P._y, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Pz, P._z, Utils::tid());
}

/***** SCALAR MULTIPLICATION OF POINT ***********************************************************************************************/

template <uint N>
__global__ void ec_multiplication_register(GPU_vector_data<N> data) {

	
	const uint NWORDS = N/32+ (N%32?1:0);
	IntegerReg<NWORDS> prime, a, var1,var2,var3,var4;
	PointReg<NWORDS> P,Q;
	uint minvp;
	Integer<N> k;
	
	// load the points P and Q from global memory 
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._x, (uint*)data.__v_Px, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._y, (uint*)data.__v_Py, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(P._z, (uint*)data.__v_Pz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(a, (uint*)data.__v_a, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(prime, (uint*)data.__v_prime, Utils::tid());
	minvp= data.__v_invp[Utils::tid()];

	// load the integer k from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(k, data.__v_k, Utils::tid());

	//Point coordinate in Montgomery form (x*R mod prime ...)
	//convertToMontg(P._x, P._x, prime, var1);
	//convertToMontg(P._y, P._y, prime, var1);
	//convertToMontg(P._z, P._z, prime, var1);
	
	scalarMul(Q,k,P,a,prime,minvp,var1,var2,var3,var4);

	//var1.set_one();
	//convertFromMontg(Q._x, Q._x, prime, minvp, var1, var2);
	//convertFromMontg(Q._y, Q._y, prime, minvp, var1, var2);
	//convertFromMontg(Q._z, Q._z, prime, minvp, var1, var2);

	// store the Point Q  in global memory
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Px, Q._x, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Py, Q._y, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Pz, Q._z, Utils::tid());
}

#endif //KERNEL_EC_ARITHMETIC_CUH_
