/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef KERNEL_FP_ARITHMETIC_CUH_
#define KERNEL_FP_ARITHMETIC_CUH_

#include "gpu_memory.cuh"
#include "integer.cuh"
#include "integer-reg.cuh"


/**************************
 * SHARED MEMORY VERSION  *
 **************************/

template<uint N>
struct SharedData3 {
	GFp<N>     _Fp;
	Integer<N> _op1;
	Integer<N> _op2;
};

/***** ADDITION IN FP (SHARED MEMORY) **************************************************************************************************/

template <uint N, uint nb_op>
__global__ void fp_addition_shared(GPU_vector_data<N> data) {

	__shared__ uint shared_integer[BLOCK_SIZE*(sizeof(SharedData3<N>) >> 2)];

	SharedData3<N>* ptr = (SharedData3<N>*) &shared_integer[Utils::tid_block()*(sizeof(SharedData3<N>) >> 2)];

	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_Fp._prime, data.__v_prime, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_op1, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_op2, data.__v_Py, Utils::tid());

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		ptr->_Fp.add_in(ptr->_op1, ptr->_op2);
	}
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, ptr->_op1, Utils::tid());
}

/***** MULTIPLICATION IN FP (SHARED MEMORY) ********************************************************************************************/

template <uint N, uint nb_op>
__global__ void fp_multiplication_shared(GPU_vector_data<N> data) {
	__shared__ uint shared_integer[BLOCK_SIZE*(sizeof(SharedData3<N>) >> 2)];
	SharedData3<N>* ptr = (SharedData3<N>*) &shared_integer[Utils::tid_block()*(sizeof(SharedData3<N>) >> 2)];

	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_Fp._prime, data.__v_prime, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_op1, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_op2, data.__v_Py, Utils::tid());
	ptr->_Fp._invp = data.__v_invp[Utils::tid()];

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		ptr->_Fp.mul(ptr->_op1, ptr->_op1, ptr->_op2);
	}
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, ptr->_op1, Utils::tid());
}



/****************************
 * REGISTER MEMORY VERSION  *
 ****************************/

/***** ADDITION IN FP (REGISTER MEMORY) **************************************************************************************************/

template <uint N, uint nb_op>
__global__ void fp_addition_register(GPU_vector_data<N> data) {

	const uint NWORDS = N/32+ (N%32?1:0);
	IntegerReg<NWORDS> prime, x,y;

	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(prime, (uint*)data.__v_prime, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(x    , (uint*)data.__v_Px   , Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(y    , (uint*)data.__v_Py   , Utils::tid());

	for(uint i =0; i < nb_op; i++) {
		addin(x,y,prime);
	}
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Px, x, Utils::tid());
}

/***** MULTIPLICATION IN FP (SHARED MEMORY) ********************************************************************************************/

template <uint N, uint nb_op>
__global__ void fp_multiplication_register(GPU_vector_data<N> data) {
	
	
	const uint NWORDS = N/32+ (N%32?1:0);
	IntegerReg<NWORDS> prime, x,y,var;
	uint minvp;

	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(prime, (uint*)data.__v_prime, Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(x    , (uint*)data.__v_Px   , Utils::tid());
	MMemory<NWORDS,NUMBER_THREADS>::globalLoad(y    , (uint*)data.__v_Py   , Utils::tid());
	minvp = (uint)data.__v_invp[Utils::tid()];


	for(uint i =0; i < nb_op; i++) {
		mul(x,x,y,prime,minvp,var);
	}
	MMemory<NWORDS,NUMBER_THREADS>::globalStore((uint*)data.__v_Px, x, Utils::tid());

}


/*************************
 * LOCAL MEMORY VERSION  *
 *************************/



/***** ADDITION IN FP (LOCAL MEMORY) **************************************************************************************************/

template <uint N, uint nb_op>
__global__ void fp_addition_local(GPU_vector_data<N> data) {


	GFp<N>  Fp;
	Integer<N> op1,op2;
	

	Memory<N,NUMBER_THREADS>::globalLoad(Fp._prime, data.__v_prime, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(op1, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(op2, data.__v_Py, Utils::tid());

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		Fp.add_in(op1, op2);
	}
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, op1, Utils::tid());
}

/***** MULTIPLICATION IN FP (LOCAL MEMORY) ********************************************************************************************/

template <uint N, uint nb_op>
__global__ void fp_multiplication_local(GPU_vector_data<N> data) {

	GFp<N>  Fp;
	Integer<N> op1,op2;	

	Memory<N,NUMBER_THREADS>::globalLoad(Fp._prime, data.__v_prime, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(op1, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(op2, data.__v_Py, Utils::tid());
	Fp._invp = data.__v_invp[Utils::tid()];

#pragma unroll
	for(uint i =0; i < nb_op; i++) {
		Fp.mul(op1, op1, op2);
	}
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, op1, Utils::tid());
}


/*******************
 * TEXTURE VERSION *
 *******************/

/***** ADDITION IN FP (TEXTURE MEMORY) **************************************************************************************************/


template <uint N, uint nb_op>
__global__ void fp_addition_texture(GPU_vector_data<N> data) {
	GFp<N,TextureMemory<0> > Fp;
	//	Integer<N,TextureMemory<0> > prime;
	Integer<N,TextureMemory<1> > op1;
	Integer<N,TextureMemory<2> > op2;
	Integer<N> op3;
	Fp.add(op3,op1,op2);
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, op3, Utils::tid());
}

template <uint N, uint nb_op>
__global__ void fp_multiplication_texture() {
	GFp<N,TextureMemory<0> > Fp;
	//	Integer<N,TextureMemory<0> > prime;
	Integer<N,TextureMemory<1> > op1;
	Integer<N,TextureMemory<2> > op2;
	Integer<N> op3;
	Fp.mul(op3,op1,op2);
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Px, op3, Utils::tid());
}


#endif //KERNEL_FP_ARITHMETIC_CUH_
